"use strict";
// TODO (jimmywarting): in the feature use conditional loading with top level await (requires 14.x)
// Node has recently added whatwg stream into core
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var __asyncValues = (this && this.__asyncValues) || function (o) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var m = o[Symbol.asyncIterator], i;
    return m ? m.call(o) : (o = typeof __values === "function" ? __values(o) : o[Symbol.iterator](), i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i);
    function verb(n) { i[n] = o[n] && function (v) { return new Promise(function (resolve, reject) { v = o[n](v), settle(resolve, reject, v.done, v.value); }); }; }
    function settle(resolve, reject, d, v) { Promise.resolve(v).then(function(v) { resolve({ value: v, done: d }); }, reject); }
};
var __await = (this && this.__await) || function (v) { return this instanceof __await ? (this.v = v, this) : new __await(v); }
var __asyncDelegator = (this && this.__asyncDelegator) || function (o) {
    var i, p;
    return i = {}, verb("next"), verb("throw", function (e) { throw e; }), verb("return"), i[Symbol.iterator] = function () { return this; }, i;
    function verb(n, f) { i[n] = o[n] ? function (v) { return (p = !p) ? { value: __await(o[n](v)), done: n === "return" } : f ? f(v) : v; } : f; }
};
var __asyncGenerator = (this && this.__asyncGenerator) || function (thisArg, _arguments, generator) {
    if (!Symbol.asyncIterator) throw new TypeError("Symbol.asyncIterator is not defined.");
    var g = generator.apply(thisArg, _arguments || []), i, q = [];
    return i = {}, verb("next"), verb("throw"), verb("return"), i[Symbol.asyncIterator] = function () { return this; }, i;
    function verb(n) { if (g[n]) i[n] = function (v) { return new Promise(function (a, b) { q.push([n, v, a, b]) > 1 || resume(n, v); }); }; }
    function resume(n, v) { try { step(g[n](v)); } catch (e) { settle(q[0][3], e); } }
    function step(r) { r.value instanceof __await ? Promise.resolve(r.value.v).then(fulfill, reject) : settle(q[0][2], r); }
    function fulfill(value) { resume("next", value); }
    function reject(value) { resume("throw", value); }
    function settle(f, v) { if (f(v), q.shift(), q.length) resume(q[0][0], q[0][1]); }
};
var _Blob_parts, _Blob_type, _Blob_size;
Object.defineProperty(exports, "__esModule", { value: true });
exports.Blob = void 0;
require("./streams.cjs");
/** @typedef {import('buffer').Blob} NodeBlob} */
// 64 KiB (same size chrome slice theirs blob into Uint8array's)
const POOL_SIZE = 65536;
/** @param {(Blob | NodeBlob | Uint8Array)[]} parts */
function toIterator(parts, clone = true) {
    return __asyncGenerator(this, arguments, function* toIterator_1() {
        for (let part of parts) {
            if ('stream' in part) {
                yield __await(yield* __asyncDelegator(__asyncValues(part.stream())));
            }
            else if (ArrayBuffer.isView(part)) {
                if (clone) {
                    let position = part.byteOffset;
                    let end = part.byteOffset + part.byteLength;
                    while (position !== end) {
                        const size = Math.min(end - position, POOL_SIZE);
                        const chunk = part.buffer.slice(position, position + size);
                        position += chunk.byteLength;
                        yield yield __await(new Uint8Array(chunk));
                    }
                }
                else {
                    yield yield __await(part);
                }
            }
            else {
                /* c8 ignore start */
                // For blobs that have arrayBuffer but no stream method (nodes buffer.Blob)
                let position = 0;
                while (position !== part.size) {
                    const chunk = part.slice(position, Math.min(part.size, position + POOL_SIZE));
                    const buffer = yield __await(chunk.arrayBuffer());
                    position += buffer.byteLength;
                    yield yield __await(new Uint8Array(buffer));
                }
                /* c8 ignore end */
            }
        }
    });
}
const _Blob = class Blob {
    /**
     * The Blob() constructor returns a new Blob object. The content
     * of the blob consists of the concatenation of the values given
     * in the parameter array.
     *
     * @param {*} blobParts
     * @param {{ type?: string }} [options]
     */
    constructor(blobParts = [], options = {}) {
        /** @type {Array.<(Blob|Uint8Array)>} */
        _Blob_parts.set(this, []);
        _Blob_type.set(this, '');
        _Blob_size.set(this, 0);
        const parts = [];
        let size = 0;
        if (typeof blobParts !== 'object') {
            throw new TypeError(`Failed to construct 'Blob': parameter 1 is not an iterable object.`);
        }
        if (typeof options !== 'object' && typeof options !== 'function') {
            throw new TypeError(`Failed to construct 'Blob': parameter 2 cannot convert to dictionary.`);
        }
        if (options === null)
            options = {};
        for (const element of blobParts) {
            let part;
            if (ArrayBuffer.isView(element)) {
                part = new Uint8Array(element.buffer.slice(element.byteOffset, element.byteOffset + element.byteLength));
            }
            else if (element instanceof ArrayBuffer) {
                part = new Uint8Array(element.slice(0));
            }
            else if (element instanceof Blob) {
                part = element;
            }
            else {
                part = new TextEncoder().encode(element);
            }
            size += ArrayBuffer.isView(part) ? part.byteLength : part.size;
            parts.push(part);
        }
        const type = options.type === undefined ? '' : String(options.type);
        __classPrivateFieldSet(this, _Blob_type, /^[\x20-\x7E]*$/.test(type) ? type : '', "f");
        __classPrivateFieldSet(this, _Blob_size, size, "f");
        __classPrivateFieldSet(this, _Blob_parts, parts, "f");
    }
    /**
     * The Blob interface's size property returns the
     * size of the Blob in bytes.
     */
    get size() {
        return __classPrivateFieldGet(this, _Blob_size, "f");
    }
    /**
     * The type property of a Blob object returns the MIME type of the file.
     */
    get type() {
        return __classPrivateFieldGet(this, _Blob_type, "f");
    }
    /**
     * The text() method in the Blob interface returns a Promise
     * that resolves with a string containing the contents of
     * the blob, interpreted as UTF-8.
     *
     * @return {Promise<string>}
     */
    text() {
        var e_1, _a;
        return __awaiter(this, void 0, void 0, function* () {
            // More optimized than using this.arrayBuffer()
            // that requires twice as much ram
            const decoder = new TextDecoder();
            let str = '';
            try {
                for (var _b = __asyncValues(toIterator(__classPrivateFieldGet(this, _Blob_parts, "f"), false)), _c; _c = yield _b.next(), !_c.done;) {
                    let part = _c.value;
                    str += decoder.decode(part, { stream: true });
                }
            }
            catch (e_1_1) { e_1 = { error: e_1_1 }; }
            finally {
                try {
                    if (_c && !_c.done && (_a = _b.return)) yield _a.call(_b);
                }
                finally { if (e_1) throw e_1.error; }
            }
            // Remaining
            str += decoder.decode();
            return str;
        });
    }
    /**
     * The arrayBuffer() method in the Blob interface returns a
     * Promise that resolves with the contents of the blob as
     * binary data contained in an ArrayBuffer.
     *
     * @return {Promise<ArrayBuffer>}
     */
    arrayBuffer() {
        var e_2, _a;
        return __awaiter(this, void 0, void 0, function* () {
            // Easier way... Just a unnecessary overhead
            // const view = new Uint8Array(this.size);
            // await this.stream().getReader({mode: 'byob'}).read(view);
            // return view.buffer;
            const data = new Uint8Array(this.size);
            let offset = 0;
            try {
                for (var _b = __asyncValues(toIterator(__classPrivateFieldGet(this, _Blob_parts, "f"), false)), _c; _c = yield _b.next(), !_c.done;) {
                    const chunk = _c.value;
                    data.set(chunk, offset);
                    offset += chunk.length;
                }
            }
            catch (e_2_1) { e_2 = { error: e_2_1 }; }
            finally {
                try {
                    if (_c && !_c.done && (_a = _b.return)) yield _a.call(_b);
                }
                finally { if (e_2) throw e_2.error; }
            }
            return data.buffer;
        });
    }
    stream() {
        const it = toIterator(__classPrivateFieldGet(this, _Blob_parts, "f"), true);
        return new ReadableStream({
            type: 'bytes',
            pull(ctrl) {
                return __awaiter(this, void 0, void 0, function* () {
                    const chunk = yield it.next();
                    chunk.done ? ctrl.close() : ctrl.enqueue(chunk.value);
                });
            }
        });
    }
    /**
     * The Blob interface's slice() method creates and returns a
     * new Blob object which contains data from a subset of the
     * blob on which it's called.
     *
     * @param {number} [start]
     * @param {number} [end]
     * @param {string} [type]
     */
    slice(start = 0, end = this.size, type = '') {
        const { size } = this;
        let relativeStart = start < 0 ? Math.max(size + start, 0) : Math.min(start, size);
        let relativeEnd = end < 0 ? Math.max(size + end, 0) : Math.min(end, size);
        const span = Math.max(relativeEnd - relativeStart, 0);
        const parts = __classPrivateFieldGet(this, _Blob_parts, "f");
        const blobParts = [];
        let added = 0;
        for (const part of parts) {
            // don't add the overflow to new blobParts
            if (added >= span) {
                break;
            }
            const size = ArrayBuffer.isView(part) ? part.byteLength : part.size;
            if (relativeStart && size <= relativeStart) {
                // Skip the beginning and change the relative
                // start & end position as we skip the unwanted parts
                relativeStart -= size;
                relativeEnd -= size;
            }
            else {
                let chunk;
                if (ArrayBuffer.isView(part)) {
                    chunk = part.subarray(relativeStart, Math.min(size, relativeEnd));
                    added += chunk.byteLength;
                }
                else {
                    chunk = part.slice(relativeStart, Math.min(size, relativeEnd));
                    added += chunk.size;
                }
                relativeEnd -= size;
                blobParts.push(chunk);
                relativeStart = 0; // All next sequential parts should start at 0
            }
        }
        const blob = new Blob([], { type: String(type).toLowerCase() });
        __classPrivateFieldSet(blob, _Blob_size, span, "f");
        __classPrivateFieldSet(blob, _Blob_parts, blobParts, "f");
        return blob;
    }
    get [(_Blob_parts = new WeakMap(), _Blob_type = new WeakMap(), _Blob_size = new WeakMap(), Symbol.toStringTag)]() {
        return 'Blob';
    }
    static [Symbol.hasInstance](object) {
        return (object &&
            typeof object === 'object' &&
            typeof object.constructor === 'function' &&
            (typeof object.stream === 'function' ||
                typeof object.arrayBuffer === 'function') &&
            /^(Blob|File)$/.test(object[Symbol.toStringTag]));
    }
};
Object.defineProperties(_Blob.prototype, {
    size: { enumerable: true },
    type: { enumerable: true },
    slice: { enumerable: true }
});
/** @type {typeof globalThis.Blob} */
exports.Blob = _Blob;
exports.default = exports.Blob;
